@@ -1,25 +1,25 @@
+import functools
+import hashlib
+import inspect
+import itertools
 import logging
-import time
-import threading
 import math
+import pickle
+import queue
 import random
 import re
+import threading
+import time
 import warnings
-import functools
-import itertools
-import hashlib
 import weakref
-import queue
-import pickle
 from collections import defaultdict
 from collections.abc import MutableMapping
 
-import six
 import redis
+import six
 
 from ._zset import ZSet
 
-
 LOGGER = logging.getLogger('fakeredis')
 REDIS_LOG_LEVELS = {
     b'LOG_DEBUG': 0,
@@ -86,6 +86,9 @@ RESTORE_INVALID_TTL_MSG = "ERR Invalid TTL value, must be >= 0"
 
 FLAG_NO_SCRIPT = 's'      # Command not allowed in scripts
 
+# This needs to be grabbed early to avoid breaking tests that mock redis.Redis.
+_ORIG_SIG = inspect.signature(redis.Redis)
+
 
 class SimpleString:
     def __init__(self, value):
@@ -240,6 +243,7 @@ class CommandItem:
     def expireat(self, value):
         self._expireat = value
         self._expireat_modified = True
+        self._modified = True  # Since redis 6.0.7
 
     def get(self, default):
         return self._value if self else default
@@ -645,6 +649,8 @@ class FakeServer:
 
 
 class FakeSocket:
+    _connection_error_class = redis.ConnectionError
+
     def __init__(self, server):
         self._server = server
         self._db = server.dbs[0]
@@ -824,7 +830,7 @@ class FakeSocket:
 
     def sendall(self, data):
         if not self._server.connected:
-            raise redis.ConnectionError(CONNECTION_ERROR_MSG)
+            raise self._connection_error_class(CONNECTION_ERROR_MSG)
         if isinstance(data, str):
             data = data.encode('ascii')
         self._parser.send(data)
@@ -1492,6 +1498,7 @@ class FakeSocket:
         xx = False
         nx = False
         keepttl = False
+        get = False
         while i < len(args):
             if casematch(args[i], b'nx'):
                 nx = True
@@ -1501,26 +1508,38 @@ class FakeSocket:
                 i += 1
             elif casematch(args[i], b'ex') and i + 1 < len(args):
                 ex = Int.decode(args[i + 1])
-                if ex <= 0:
+                if ex <= 0 or (self._db.time + ex) * 1000 >= 2**63:
                     raise SimpleError(INVALID_EXPIRE_MSG.format('set'))
                 i += 2
             elif casematch(args[i], b'px') and i + 1 < len(args):
                 px = Int.decode(args[i + 1])
-                if px <= 0:
+                if px <= 0 or self._db.time * 1000 + px >= 2**63:
                     raise SimpleError(INVALID_EXPIRE_MSG.format('set'))
                 i += 2
             elif casematch(args[i], b'keepttl'):
                 keepttl = True
                 i += 1
+            elif casematch(args[i], b'get'):
+                get = True
+                i += 1
             else:
                 raise SimpleError(SYNTAX_ERROR_MSG)
         if (xx and nx) or ((px is not None) + (ex is not None) + keepttl > 1):
             raise SimpleError(SYNTAX_ERROR_MSG)
+        if nx and get:
+            # The command docs say this is allowed from Redis 7.0.
+            raise SimpleError(SYNTAX_ERROR_MSG)
+
+        old_value = None
+        if get:
+            if key.value is not None and type(key.value) is not bytes:
+                raise SimpleError(WRONGTYPE_MSG)
+            old_value = key.value
 
         if nx and key:
-            return None
+            return old_value
         if xx and not key:
-            return None
+            return old_value
         if not keepttl:
             key.value = value
         else:
@@ -1529,11 +1548,11 @@ class FakeSocket:
             key.expireat = self._db.time + ex
         if px is not None:
             key.expireat = self._db.time + px / 1000.0
-        return OK
+        return OK if not get else old_value
 
     @command((Key(), Int, bytes))
     def setex(self, key, seconds, value):
-        if seconds <= 0:
+        if seconds <= 0 or (self._db.time + seconds) * 1000 >= 2**63:
             raise SimpleError(INVALID_EXPIRE_MSG.format('setex'))
         key.value = value
         key.expireat = self._db.time + seconds
@@ -1541,7 +1560,7 @@ class FakeSocket:
 
     @command((Key(), Int, bytes))
     def psetex(self, key, ms, value):
-        if ms <= 0:
+        if ms <= 0 or self._db.time * 1000 + ms >= 2**63:
             raise SimpleError(INVALID_EXPIRE_MSG.format('psetex'))
         key.value = value
         key.expireat = self._db.time + ms / 1000.0
@@ -1752,14 +1771,38 @@ class FakeSocket:
     def llen(self, key):
         return len(key.value)
 
-    @command((Key(list),))
-    def lpop(self, key):
-        try:
-            ret = key.value.pop(0)
-            key.updated()
-            return ret
-        except IndexError:
+    def _list_pop(self, get_slice, key, *args):
+        """Implements lpop and rpop.
+
+        `get_slice` must take a count and return a slice expression for the
+        range to pop.
+        """
+        # This implementation is somewhat contorted to match the odd
+        # behaviours described in https://github.com/redis/redis/issues/9680.
+        count = 1
+        if len(args) > 1:
+            raise SimpleError(SYNTAX_ERROR_MSG)
+        elif len(args) == 1:
+            count = args[0]
+            if count < 0:
+                raise SimpleError(INDEX_ERROR_MSG)
+            elif count == 0:
+                return None
+        if not key:
             return None
+        elif type(key.value) != list:
+            raise SimpleError(WRONGTYPE_MSG)
+        slc = get_slice(count)
+        ret = key.value[slc]
+        del key.value[slc]
+        key.updated()
+        if not args:
+            ret = ret[0]
+        return ret
+
+    @command((Key(),), (Int(),))
+    def lpop(self, key, *args):
+        return self._list_pop(lambda count: slice(None, count), key, *args)
 
     @command((Key(list), bytes), (bytes,))
     def lpush(self, key, *values):
@@ -1824,14 +1867,9 @@ class FakeSocket:
                 key.update(new_value)
         return OK
 
-    @command((Key(list),))
-    def rpop(self, key):
-        try:
-            ret = key.value.pop()
-            key.updated()
-            return ret
-        except IndexError:
-            return None
+    @command((Key(),), (Int(),))
+    def rpop(self, key, *args):
+        return self._list_pop(lambda count: slice(None, -count - 1, -1), key, *args)
 
     @command((Key(list, None), Key(list)))
     def rpoplpush(self, src, dst):
@@ -1900,13 +1938,11 @@ class FakeSocket:
     def sdiffstore(self, dst, *keys):
         return self._setop(lambda a, b: a - b, False, dst, *keys)
 
-    # The following keys can't be marked as sets because of the
-    # stop_if_empty early-out.
-    @command((Key(set),), (Key(),))
+    @command((Key(set),), (Key(set),))
     def sinter(self, *keys):
         return self._setop(lambda a, b: a & b, True, None, *keys)
 
-    @command((Key(), Key(set)), (Key(),))
+    @command((Key(), Key(set)), (Key(set),))
     def sinterstore(self, dst, *keys):
         return self._setop(lambda a, b: a & b, True, dst, *keys)
 
@@ -1935,7 +1971,7 @@ class FakeSocket:
         if count is None:
             if not key.value:
                 return None
-            item = random.sample(key.value, 1)[0]
+            item = random.sample(list(key.value), 1)[0]
             key.value.remove(item)
             key.updated()
             return item
@@ -1954,10 +1990,10 @@ class FakeSocket:
             if not key.value:
                 return None
             else:
-                return random.sample(key.value, 1)[0]
+                return random.sample(list(key.value), 1)[0]
         elif count >= 0:
             count = min(count, len(key.value))
-            return random.sample(key.value, count)
+            return random.sample(list(key.value), count)
         else:
             items = list(key.value)
             return [random.choice(items) for _ in range(-count)]
@@ -2129,14 +2165,17 @@ class FakeSocket:
 
     def _zrange(self, key, start, stop, reverse, *args):
         zset = key.value
-        # TODO: does redis allow multiple WITHSCORES?
-        if len(args) > 1 or (args and not casematch(args[0], b'withscores')):
-            raise SimpleError(SYNTAX_ERROR_MSG)
+        withscores = False
+        for arg in args:
+            if casematch(arg, b'withscores'):
+                withscores = True
+            else:
+                raise SimpleError(SYNTAX_ERROR_MSG)
         start, stop = self._fix_range(start, stop, len(zset))
         if reverse:
             start, stop = len(zset) - stop, len(zset) - start
         items = zset.islice_score(start, stop, reverse)
-        items = self._apply_withscores(items, bool(args))
+        items = self._apply_withscores(items, withscores)
         return items
 
     @command((Key(ZSet), Int, Int), (bytes,))
@@ -2353,8 +2392,10 @@ class FakeSocket:
     # Server commands
     # TODO: lots
 
-    @command((), flags='s')
-    def bgsave(self):
+    @command((), (bytes,), flags='s')
+    def bgsave(self, *args):
+        if len(args) > 1 or (len(args) == 1 and not casematch(args[0], b'schedule')):
+            raise SimpleError(SYNTAX_ERROR_MSG)
         self._server.lastsave = int(time.time())
         return BGSAVE_STARTED
 
@@ -2491,7 +2532,7 @@ class FakeSocket:
 
     @command((bytes, Int), (bytes,), flags='s')
     def eval(self, script, numkeys, *keys_and_args):
-        from lupa import LuaRuntime, LuaError, as_attrgetter
+        from lupa import LuaError, LuaRuntime, as_attrgetter
 
         if numkeys > len(keys_and_args):
             raise SimpleError(TOO_MANY_KEYS_MSG)
@@ -2558,7 +2599,7 @@ class FakeSocket:
         elif casematch(subcmd, b'exists'):
             return [int(sha1 in self._server.script_cache) for sha1 in args]
         elif casematch(subcmd, b'flush'):
-            if len(args) != 0:
+            if len(args) > 1 or (len(args) == 1 and casenorm(args[0]) not in {b'sync', b'async'}):
                 raise SimpleError(BAD_SUBCOMMAND_MSG.format('SCRIPT'))
             self._server.script_cache = {}
             return OK
@@ -2716,7 +2757,7 @@ class FakeConnection(redis.Connection):
         else:
             return response
 
-    def read_response(self):
+    def read_response(self, disable_decoding=False):
         if not self._server.connected:
             try:
                 response = self._sock.responses.get_nowait()
@@ -2726,7 +2767,10 @@ class FakeConnection(redis.Connection):
             response = self._sock.responses.get()
         if isinstance(response, redis.ResponseError):
             raise response
-        return self._decode(response)
+        if disable_decoding:
+            return response
+        else:
+            return self._decode(response)
 
     def repr_pieces(self):
         pieces = [
@@ -2739,73 +2783,63 @@ class FakeConnection(redis.Connection):
 
 
 class FakeRedisMixin:
-    def __init__(self, host='localhost', port=6379,
-                 db=0, password=None, socket_timeout=None,
-                 socket_connect_timeout=None,
-                 socket_keepalive=None, socket_keepalive_options=None,
-                 connection_pool=None, unix_socket_path=None,
-                 encoding='utf-8', encoding_errors='strict',
-                 charset=None, errors=None,
-                 decode_responses=False, retry_on_timeout=False,
-                 ssl=False, ssl_keyfile=None, ssl_certfile=None,
-                 ssl_cert_reqs=None, ssl_ca_certs=None,
-                 max_connections=None, server=None,
-                 connected=True):
-        if not connection_pool:
+    def __init__(self, *args, server=None, connected=True, **kwargs):
+        # Interpret the positional and keyword arguments according to the
+        # version of redis in use.
+        bound = _ORIG_SIG.bind(*args, **kwargs)
+        bound.apply_defaults()
+        if not bound.arguments['connection_pool']:
+            charset = bound.arguments['charset']
+            errors = bound.arguments['errors']
             # Adapted from redis-py
             if charset is not None:
                 warnings.warn(DeprecationWarning(
                     '"charset" is deprecated. Use "encoding" instead'))
-                encoding = charset
+                bound.arguments['encoding'] = charset
             if errors is not None:
                 warnings.warn(DeprecationWarning(
                     '"errors" is deprecated. Use "encoding_errors" instead'))
-                encoding_errors = errors
+                bound.arguments['encoding_errors'] = errors
 
             if server is None:
                 server = FakeServer()
                 server.connected = connected
             kwargs = {
-                'db': db,
-                'password': password,
-                'encoding': encoding,
-                'encoding_errors': encoding_errors,
-                'decode_responses': decode_responses,
-                'max_connections': max_connections,
                 'connection_class': FakeConnection,
                 'server': server
             }
-            connection_pool = redis.connection.ConnectionPool(**kwargs)
-        # These need to be passed by name due to
-        # https://github.com/andymccurdy/redis-py/issues/1276
-        super().__init__(
-            host=host, port=port, db=db, password=password, socket_timeout=socket_timeout,
-            socket_connect_timeout=socket_connect_timeout,
-            socket_keepalive=socket_keepalive,
-            socket_keepalive_options=socket_keepalive_options,
-            connection_pool=connection_pool,
-            unix_socket_path=unix_socket_path,
-            encoding=encoding, encoding_errors=encoding_errors,
-            charset=charset, errors=errors,
-            decode_responses=decode_responses, retry_on_timeout=retry_on_timeout,
-            ssl=ssl, ssl_keyfile=ssl_keyfile, ssl_certfile=ssl_certfile,
-            ssl_cert_reqs=ssl_cert_reqs, ssl_ca_certs=ssl_ca_certs,
-            max_connections=max_connections)
+            conn_pool_args = [
+                'db',
+                'username',
+                'password',
+                'socket_timeout',
+                'encoding',
+                'encoding_errors',
+                'decode_responses',
+                'retry_on_timeout',
+                'max_connections',
+                'health_check_interval',
+                'client_name'
+            ]
+            for arg in conn_pool_args:
+                if arg in bound.arguments:
+                    kwargs[arg] = bound.arguments[arg]
+            bound.arguments['connection_pool'] = redis.connection.ConnectionPool(**kwargs)
+        super().__init__(*bound.args, **bound.kwargs)
 
     @classmethod
-    def from_url(cls, url, db=None, **kwargs):
+    def from_url(cls, *args, **kwargs):
         server = kwargs.pop('server', None)
         if server is None:
             server = FakeServer()
-        self = super().from_url(url, db, **kwargs)
+        pool = redis.ConnectionPool.from_url(*args, **kwargs)
         # Now override how it creates connections
-        pool = self.connection_pool
         pool.connection_class = FakeConnection
         pool.connection_kwargs['server'] = server
-        for key in ['password', 'host', 'port', 'path']:
-            if key in pool.connection_kwargs:
-                del pool.connection_kwargs[key]
-        return self
+        # FakeConnection cannot handle the path kwarg (present when from_url
+        # is called with a unix socket)
+        pool.connection_kwargs.pop('path', None)
+        return cls(connection_pool=pool)
 
 
 class FakeStrictRedis(FakeRedisMixin, redis.StrictRedis):
